#ifndef NETWORK_H
#define NETWORK_H

#include "player.h"

// If this line is enable, every frame will have a sync test
//  this is not needed in normal games. Normal is like 1 sync in 100
//  frames. You can enable this if you have a lot of desyncs on a certain
//  game.
// Remember: both client and server have to be compiled with this
//  option enabled to make it to work. If one of the two has it disabled
//  nothing will happen.
//#define ENABLE_NETWORK_SYNC_EVERY_FRAME

// In theory sending 1 of the 2 seeds is enough to check for desyncs
//   so in theory, this next define can be left off.
//#define NETWORK_SEND_DOUBLE_SEED

// How many clients can we have? Like.. MAX_PLAYERS - 1 is the amount of
//  players that can really play.. so.. a max of 4 spectators.. gives us..
//  MAX_PLAYERS + 3
#define MAX_CLIENTS (MAX_PLAYERS + 3)

#define NETWORK_DISCOVER_PORT 3978
#define NETWORK_DEFAULT_PORT 3979

// This is the struct used by both client and server
//  some fields will be empty on the client (like game_password) by default
//  and only filled with data a player enters.
typedef struct NetworkGameInfo {
	char server_name[40];			// Server name
	uint16 server_revision;		// The SVN version number the server is using (e.g.: '304')
														//  It even shows a SVN version in release-version, so
														//  it is easy to compare if a server is of the correct version
	byte server_lang;					// Language of the server (we should make a nice table for this)
	char server_password[20];	// On the server: the game password, on the client: != "" if server has password
	byte clients_max;					// Max clients allowed on server
	byte clients_on;					// Current count of clients on server
	byte specators_on;				// How many spectators do we have?
	uint16 game_date;					// Current date
	char map_name[40];				// Map which is played ["random" for a randomized map]
	uint16 map_width;					// Map width
	uint16 map_height;				// Map height
	byte map_set;							// Graphical set
} NetworkGameInfo;

typedef struct NetworkPlayerInfo {
	uint64 player_money;			// How much money the client has
	char company_name[32];		// Name of the company
	uint64 company_value;			// The current value of the company
	char password[20];				// The password for the player
} NetworkPlayerInfo;

typedef struct NetworkClientInfo {
	char client_name[40];			// Name of the client
	byte client_lang;					// The language of the client
	byte client_playas;				// As which player is this client playing
	char client_ip[15];				// IP-address of the client (so he can be banned)
	uint16 join_date;					// Gamedate the player has joined
} NetworkClientInfo;

typedef struct NetworkGameList {
	NetworkGameInfo item;
	uint32 ip;
	uint16 port;
	struct NetworkGameList *_next;
} NetworkGameList;

NetworkGameList *_network_game_list;

NetworkGameInfo _network_game_info;
NetworkPlayerInfo _network_player_info[MAX_PLAYERS];
NetworkClientInfo _network_client_info[MAX_CLIENTS];

char _network_player_name[50];
char _network_game_password[20];

void ParseConnectionString(const byte **player, const byte **port, byte *connection_string);

#endif /* NETWORK_H */
