#ifndef NETWORK_CORE_H
#define NETWORK_CORE_H

// Network stuff has many things that needs to be included
//  by default. All those things are in this file.
// This file should only be included in network.c. It is just there to make
//  network.c a bit less messy and unreadable.

// =============================
// Include standard stuff per OS

// Windows stuff
#if defined(WIN32)
#ifdef _XBOX
#	include <xtl.h>
#	include <winsockx.h>
#else
#	include <windows.h>
#	include <winsock.h>
#	pragma comment (lib, "ws2_32.lib")
#endif
#	define ENABLE_NETWORK // On windows, the network is always enabled
#	define GET_LAST_ERROR() WSAGetLastError()
#	define EWOULDBLOCK WSAEWOULDBLOCK
// Windows has some different names for some types..
typedef SSIZE_T ssize_t;
typedef unsigned long in_addr_t;
#endif // WIN32

// UNIX stuff
#if defined(UNIX)
#	define SOCKET int
#	define INVALID_SOCKET -1
#	if !defined(__MORPHOS__) && !defined(__AMIGA__)
#		define ioctlsocket ioctl
#		define closesocket close
#		define GET_LAST_ERROR() (errno)
#	endif
// Need this for FIONREAD on solaris
#	define BSD_COMP

// Includes needed for UNIX-like systems
#	include <unistd.h>
#	include <sys/ioctl.h>
#	include <sys/socket.h>
#	include <netinet/in.h>
#	include <netinet/tcp.h>
#	include <arpa/inet.h>
#	include <errno.h>
#	include <sys/time.h>
#	include <netdb.h>
#endif // UNIX

// MorphOS and Amiga stuff
#if defined(__MORPHOS__) || defined(__AMIGA__)
#	include <exec/types.h>
#	include <proto/exec.h>		// required for Open/CloseLibrary()
#	if defined(__MORPHOS__)
#		include <sys/filio.h> 	// FION#? defines
#	else // __AMIGA__
#		include	<proto/socket.h>
#	endif

// Make the names compatible
#	define closesocket(s) CloseSocket(s)
#	define GET_LAST_ERROR() Errno()
#	define ioctlsocket(s,request,status) IoctlSocket((LONG)s,(ULONG)request,(char*)status)

	struct Library *SocketBase = NULL;
#endif // __MORPHOS__ || __AMIGA__

#endif // NETWORK_CORE_H
