#include "stdafx.h"
#include "ttd.h"
#include "hal.h"
#include <time.h>
#include <xtl.h>

#if defined(__linux__)
#include <sys/statvfs.h>
#endif


#if defined(WITH_SDL)
#include <SDL.h>
#endif

#ifdef __MORPHOS__
#include <exec/types.h>
ULONG __stack = (1024*1024)*2; // maybe not that much is needed actually ;)
#endif /* __MORPHOS__ */

static char *_fios_path;
static char *_fios_save_path;
static char *_fios_scn_path;
static FiosItem *_fios_items;
static int _fios_count, _fios_alloc;

extern HalMusicDriver _sdl_midi_music_driver;

static FiosItem *FiosAlloc()
{
	if (_fios_count == _fios_alloc) {
		_fios_alloc += 256;
		_fios_items = realloc(_fios_items, _fios_alloc * sizeof(FiosItem));
	}
	return &_fios_items[_fios_count++];
}

int compare_FiosItems (const void *a, const void *b) {
	const FiosItem *da = (const FiosItem *) a;
	const FiosItem *db = (const FiosItem *) b;
	int r;

	if (_savegame_sort_order < 2) // sort by date
    r = da->mtime < db->mtime ? -1 : 1;
	else
		r = strcmp(da->title[0] ? da->title : da->name, db->title[0] ? db->title : db->name);

	if (_savegame_sort_order & 1) r = -r;
	return r;
}

static HANDLE MyFindFirstFile(char *path, char *file, WIN32_FIND_DATA *fd)
{
	char paths[MAX_PATH];

	sprintf(paths, "%s\\%s", path, file);
	return FindFirstFile(paths, fd);
}

// Get a list of savegames
FiosItem *FiosGetSavegameList(int *num, int mode)
{
	WIN32_FIND_DATA fd;
	HANDLE h;
	FiosItem *fios;
	int sort_start;
	char buf[MAX_PATH];

	if (_fios_save_path == NULL) {
		_fios_save_path = malloc(MAX_PATH);
		strcpy(_fios_save_path, _path.save_dir);
	}

	if (_game_mode == GM_EDITOR)
		_fios_path = _fios_scn_path;
	else
		_fios_path = _fios_save_path;

	// Parent directory, only if not of the type C:\.
	ShowInfoF("Save path: %s fios path: %s\n",_path.save_dir,_fios_path);
	if (_fios_path[3] != 0 && stricmp(_path.save_dir,_fios_path)) {
		fios = FiosAlloc();
		fios->type = FIOS_TYPE_PARENT;
		strcpy(fios->title, ".. (Parent directory)");
	}


	// Show subdirectories first
	h = MyFindFirstFile(_fios_path, "*", &fd);
	if (h != INVALID_HANDLE_VALUE) {
		do {
			if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY &&
					!(fd.cFileName[0]=='.' && (fd.cFileName[1]==0 || fd.cFileName[1]=='.' && fd.cFileName[2]==0))) {
				fios = FiosAlloc();
				fios->type = FIOS_TYPE_DIR;
				strcpy(fios->name, fd.cFileName);
				sprintf(fios->title, "\\%s (Directory)", fd.cFileName);
			}
		} while (FindNextFile(h, &fd));
		FindClose(h);
	}

	{
		/* XXX ugly global variables ... */
		byte order = _savegame_sort_order;
		_savegame_sort_order = 2; // sort ascending by name
		qsort(_fios_items, _fios_count, sizeof(FiosItem), compare_FiosItems);
		_savegame_sort_order = order;
	}

	// this is where to start sorting
	sort_start = _fios_count;

	/*	Show savegame files
	 *	.SAV	OpenTTD saved game
	 *	.SS1	Transport Tycoon Deluxe preset game
	 *	.SV1	Transport Tycoon Deluxe (Patch) saved game
	 *	.SV2	Transport Tycoon Deluxe (Patch) saved 2-player game
	 */
	h = MyFindFirstFile(_fios_path, "*.*", &fd);
	if (h != INVALID_HANDLE_VALUE) {
		do {
			if (!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
				char *t = strrchr(fd.cFileName, '.');
				if (t && !stricmp(t, ".SAV")) { // OpenTTD
					fios = FiosAlloc();
					fios->mtime = *(uint64*)&fd.ftLastWriteTime;
					sprintf(buf, "%s\\%s", _fios_path, fd.cFileName);
					fios->type = FIOS_TYPE_FILE;
					fios->title[0] = 0;
					ttd_strlcpy(fios->name, fd.cFileName, strlen(fd.cFileName)-3);
				}	else if (mode == SLD_LOAD_GAME || mode == SLD_LOAD_SCENARIO) {
					int ext = 0; // start of savegame extensions in _old_extensions[]
					if (t && ((ext++, !stricmp(t, ".SS1")) || (ext++, !stricmp(t, ".SV1")) || (ext++, !stricmp(t, ".SV2"))) ) { // TTDLX(Patch)
						fios = FiosAlloc();
						fios->old_extension = ext-1;
						fios->mtime = *(uint64*)&fd.ftLastWriteTime;
						sprintf(buf, "%s\\%s", _fios_path, fd.cFileName);
						fios->type = FIOS_TYPE_OLDFILE;
						GetOldSaveGameName(fios->title, buf);
						ttd_strlcpy(fios->name, fd.cFileName, strlen(fd.cFileName)-3);
					}
				}
			}
		} while (FindNextFile(h, &fd));
		FindClose(h);
	}

	qsort(_fios_items + sort_start, _fios_count - sort_start, sizeof(FiosItem), compare_FiosItems);

	*num = _fios_count;
	return _fios_items;
}

// Get a list of scenarios
FiosItem *FiosGetScenarioList(int *num, int mode)
{
	FiosItem *fios;
	WIN32_FIND_DATA fd;
	HANDLE h;
	int sort_start;
	char buf[MAX_PATH];

	if (mode == SLD_NEW_GAME || _fios_scn_path == NULL) {
		if (_fios_scn_path == NULL)
			_fios_scn_path = malloc(MAX_PATH);
		strcpy(_fios_scn_path, _path.scenario_dir);
	}

	_fios_path = _fios_scn_path;

	// Parent directory, only if not of the type C:\.
	ShowInfoF("Scenario path: %s fios path: %s\n",_path.scenario_dir,_fios_path);
	if (_fios_path[3] != 0 && mode != SLD_NEW_GAME && stricmp(_path.scenario_dir,_fios_path)) {
		fios = FiosAlloc();
		fios->type = FIOS_TYPE_PARENT;
		strcpy(fios->title, ".. (Parent directory)");
	}

	// Show subdirectories first
	h = MyFindFirstFile(_fios_scn_path, "*", &fd);
	if (h != INVALID_HANDLE_VALUE && mode != SLD_NEW_GAME) {
		do {
			if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY &&
					!(fd.cFileName[0]=='.' && (fd.cFileName[1]==0 || fd.cFileName[1]=='.' && fd.cFileName[2]==0))) {
				fios = FiosAlloc();
				fios->type = FIOS_TYPE_DIR;
				strcpy(fios->name, fd.cFileName);
				sprintf(fios->title, "\\%s (Directory)", fd.cFileName);
			}
		} while (FindNextFile(h, &fd));
		FindClose(h);
	}

	// this is where to start sorting
	sort_start = _fios_count;

	/*	Show scenario files
	 *	.SCN	OpenTTD style scenario file
	 *	.SV0	Transport Tycoon Deluxe (Patch) scenario
	 *	.SS0	Transport Tycoon Deluxe preset scenario
	 */
	h = MyFindFirstFile(_fios_scn_path, "*.*", &fd);
	if (h != INVALID_HANDLE_VALUE) {
		do {
			if (!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
				char *t = strrchr(fd.cFileName, '.');
				if (t && !stricmp(t, ".SCN")) { // OpenTTD
					fios = FiosAlloc();
					fios->mtime = *(uint64*)&fd.ftLastWriteTime;
					sprintf(buf, "%s\\%s", _fios_path, fd.cFileName);
					fios->type = FIOS_TYPE_SCENARIO;
					fios->title[0] = 0;
					ttd_strlcpy(fios->name, fd.cFileName, strlen(fd.cFileName)-3);
				}	else if (mode == SLD_LOAD_GAME || mode == SLD_LOAD_SCENARIO || mode == SLD_NEW_GAME) {
					int ext = 3; // start of scenario extensions in _old_extensions[]
					if (t && ((ext++, !stricmp(t, ".SV0")) || (ext++, !stricmp(t, ".SS0"))) ) { // TTDLX(Patch)
						fios = FiosAlloc();
						fios->old_extension = ext-1;
						fios->mtime = *(uint64*)&fd.ftLastWriteTime;
						sprintf(buf, "%s\\%s", _fios_path, fd.cFileName);
						fios->type = FIOS_TYPE_OLD_SCENARIO;
						GetOldScenarioGameName(fios->title, buf);
						ttd_strlcpy(fios->name, fd.cFileName, strlen(fd.cFileName)-3);
					}
				}
			}
		} while (FindNextFile(h, &fd));
		FindClose(h);
	}

	qsort(_fios_items + sort_start, _fios_count - sort_start, sizeof(FiosItem), compare_FiosItems);

	*num = _fios_count;
	return _fios_items;
}

// Free the list of savegames
void FiosFreeSavegameList()
{
	free(_fios_items);
	_fios_items = NULL;
	_fios_alloc = _fios_count = 0;
}

// Browse to
char *FiosBrowseTo(const FiosItem *item)
{
	char *path = _fios_path;
	char *s;

	switch(item->type) {
	case FIOS_TYPE_DRIVE:
		sprintf(path, "%c:\\", item->title[0]);
		break;

	case FIOS_TYPE_PARENT:
		// Skip drive part
		path += 3;
		s = path;
		while (*path) {
			if (*path== '\\')
				s = path;
			path++;
		}
		*s = 0;
		break;

	case FIOS_TYPE_DIR:
		// Scan to end
		while (*++path);
		// Add backslash?
		if (path[-1] != '\\') *path++ = '\\';

		strcpy(path, item->name);
		break;

	case FIOS_TYPE_FILE:
		FiosMakeSavegameName(str_buffr, item->name);
		return str_buffr;

	case FIOS_TYPE_OLDFILE:
		sprintf(str_buffr, "%s\\%s.%s", _fios_path, item->name, _old_extensions[item->old_extension]);
		return str_buffr;

	case FIOS_TYPE_SCENARIO:
		sprintf(str_buffr, "%s\\%s.scn", path, item->name);
		return str_buffr;
	case FIOS_TYPE_OLD_SCENARIO:
		sprintf(str_buffr, "%s\\%s.%s", path, item->name, _old_extensions[item->old_extension]);
		return str_buffr;
	}

	return NULL;
}

// Get descriptive texts.
// Returns a path as well as a
//  string describing the path.
StringID FiosGetDescText(const char **path)
{
	char root[4];
	DWORD spc, bps, nfc, tnc;
	*path = _fios_path;

	root[0] = _fios_path[0];
	root[1] = ':';
	root[2] = '\\';
	root[3] = 0;
/*	if (GetDiskFreeSpace(root, &spc, &bps, &nfc, &tnc)) {
		uint32 tot = ((spc*bps)*(uint64)nfc) >> 20;
		SET_DPARAM32(0, tot);
		return STR_4005_BYTES_FREE;
	} else {*/
		return STR_4006_UNABLE_TO_READ_DRIVE;
	//}
}

void FiosMakeSavegameName(char *buf, const char *name)
{   char *t;
	while(t=strchr(name,',')) {
	  strcpy(t,t+1);
	}
	if(_game_mode==GM_EDITOR)
		sprintf(buf, "%s\\%s.scn", _fios_path, name);
	else
		sprintf(buf, "%s\\%s.sav", _fios_path, name);
}

void FiosDelete(const char *name)
{
	char *path = str_buffr;
	FiosMakeSavegameName(path, name);
	DeleteFile(path);
}

const DriverDesc _video_driver_descs[] = {
	{"null",	"Null Video Driver",	&_null_video_driver,	0},
#if defined(WITH_SDL)
	{ "sdl",	"SDL Video Driver",		&_sdl_video_driver,		1},
#endif
	{ "dedicated", "Dedicated Video Driver", &_dedicated_video_driver, 0},
	{ NULL,		NULL,									NULL,									0}
};

const DriverDesc _sound_driver_descs[] = {
	{"null",	"Null Sound Driver",	&_null_sound_driver,		0},
#if defined(WITH_SDL)
	//{ "sdl",	"SDL Sound Driver",		&_sdl_sound_driver,			1},
#endif
	{ NULL,		NULL,									NULL,										0}
};

#if defined(__APPLE__)
#define EXTMIDI_PRI 2
#else
#define EXTMIDI_PRI 0
#endif

const DriverDesc _music_driver_descs[] = {
#ifndef __BEOS__
#ifndef __MORPHOS__
// MorphOS have no music support
#endif
#endif
#ifdef __BEOS__
	{ "bemidi",	"BeOS MIDI Driver",			&_bemidi_music_driver,	1},
#endif
	{"sdl_midi",	"SDL MIDI Driver",	&_sdl_midi_music_driver,	1},
	{   "null",	"Null Music Driver",		&_null_music_driver,		1},
	{     NULL,	NULL,										NULL,										0}
};

/* GetOSVersion returns the minimal required version of OS to be able to use that driver.
	 Not needed for *nix. */
byte GetOSVersion()
{
	return 2;  // any arbitrary number bigger then 0
				// numbers lower than 2 breaks default music selection on mac
}

bool FileExists(const char *filename)
{
	HANDLE hand = CreateFile(filename, 0, 0, NULL, OPEN_EXISTING, 0, NULL);
	if (hand == INVALID_HANDLE_VALUE) return false;
	CloseHandle(hand);
	return true;
}

static int LanguageCompareFunc(const void *a, const void *b)
{
	return strcmp(*(const char* const *)a, *(const char* const *)b);
}

int GetLanguageList(char **languages, int max)
{
	HANDLE hand;
	int num = 0;
	char filedir[MAX_PATH];
	WIN32_FIND_DATA fd;
	sprintf(filedir, "%s*.lng", _path.lang_dir);

	hand = FindFirstFile(filedir, &fd);
	if (hand != INVALID_HANDLE_VALUE) {
		do {
			if (!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
				languages[num++] = strdup(fd.cFileName);
				if (num == max) break;
			}
		} while (FindNextFile(hand, &fd));
		FindClose(hand);
	}

	qsort(languages, num, sizeof(char*), LanguageCompareFunc);
	return num;
}

#if 0
int GetLanguageList(char **languages, int max)
{
//	DIR *dir;
//	struct dirent *dirent;
	HANDLE hfind;
	WIN32_FIND_DATA wfd;

	int num = 0;
	char buf[200];

	sprintf(buf,"%s*.lng",_path.lang_dir);
	hfind = FindFirstFile(buf,&wfd);
	if (hfind != INVALID_HANDLE_VALUE) {
		do {
			languages[num++] = strdup(wfd.cFileName);
			if (num == max) break;
        } while( FindNextFile( hfind, &wfd ) );

        // Close the find handle.
        FindClose( hfind );
	}
#if 0
	dir = opendir(_path.lang_dir);
	if (dir != NULL) {
		while ((dirent = readdir(dir))) {
			char *t = strrchr(dirent->d_name, '.');
			if (t && !strcmp(t, ".lng")) {
				languages[num++] = strdup(dirent->d_name);
				if (num == max) break;
			}
		}
		closedir(dir);
	}

#endif
	qsort(languages, num, sizeof(char*), LanguageCompareFunc);
	return num;
}


#endif
FILE *logfile;
float xoffset,yoffset,xstretch,ystretch;

void ShowInfo(const char *str)
{
	fputs(str,logfile);
}



void ShowOSErrorBox(const char *buf)
{
	fprintf(logfile, "Error: %s\n", buf);
	fflush(logfile);
}

int CDECL main(int argc, char* argv[])
{
	char buf[200];
	FILE *screensettings;
	logfile=fopen("D:\\openttd.txt","w+");
	screensettings=fopen("D:\\xottd.cfg","r+");
	fgets(buf,199,screensettings);
	sscanf(buf,"%f %f %f %f",&xoffset,&yoffset,&xstretch,&ystretch);
	fclose(screensettings);
	ShowOSErrorBox("Starting ttd");
	// change the working directory to enable doubleclicking in UIs
#if defined(__BEOS__)
	ChangeWorkingDirectory(argv[0]);
#endif
#if defined(__linux__)
	ChangeWorkingDirectory(argv[0]);
#endif

	_random_seeds[0][1] = _random_seeds[0][0] = time(NULL);

	return ttd_main(argc, argv);
}

void DeterminePaths()
{

	_path.game_data_dir = malloc( MAX_PATH );
	strcpy(_path.game_data_dir, "D:\\");

	_path.personal_dir = malloc( MAX_PATH );
	strcpy(_path.personal_dir, "D:\\");


	_path.save_dir = str_fmt("D:\\save");
	_path.autosave_dir = str_fmt("D:\\save\\autosave");
	_path.scenario_dir = str_fmt("D:\\scenario");
	_path.gm_dir = str_fmt("D:\\gm\\");
	_path.data_dir = str_fmt("D:\\data\\");
	_path.lang_dir = str_fmt("D:\\lang\\");

	_config_file = str_fmt("D:\\openttd.cfg");

	// create necessary folders
}

